$(document).ready(function(){var a=function(a,b){return a=a.replace(/,/g,""),b=b.replace(/,/g,""),"%"==a.slice(-1)&&(a=a.slice(0,-1),b=b.slice(0,-1)),isFinite(parseFloat(a))&&(a=parseFloat(a),b=parseFloat(b)),a<b?-1:a>b?1:0};jQuery.fn.dataTableExt.oSort["sort-column-asc"]=function(b,c){return"--"==b?1:"--"==c?-1:a(b,c)},jQuery.fn.dataTableExt.oSort["sort-column-desc"]=function(b,c){return"--"==b?1:"--"==c?-1:a(c,b)};var b=$(".blox_table");for(i=0;i<b.length;i++)$(b[i]).DataTable({columnDefs:[{type:"sort-column",targets:"_all"}],lengthMenu:[[10,25,50,100,250,500,1e3,-1],[10,25,50,100,250,500,1e3,"All"]],paging:!$(b[i]).hasClass("dt-no-paging"),searching:!$(b[i]).hasClass("dt-no-searching"),info:!$(b[i]).hasClass("dt-no-info")})});